/*----------------------------------------------------------------------- 
 * 
 * File Name: ApplySearch.c
 *
 * Author: Brown, D. A.
 * 
 * Revision: $Id: ApplySearch.c,v 1.13 2001/05/30 22:10:28 duncan Exp $
 * 
 *-----------------------------------------------------------------------
 */

#include "ApplySearch.h"

NRCSID( APPLYSEARCHC, PKG "$Id: ApplySearch.c,v 1.13 2001/05/30 22:10:28 duncan Exp $" );


void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams
              )
{
  InspiralSearchParams         *params       = NULL;
  FindChirpMasterParams        *masterParams = NULL;
  FindChirpSlaveParams         *slaveParams  = NULL;
  InspiralEvent                *eventList    = NULL;
  INT4                          myRank;


  INITSTATUS( status, "LALApplySearch", APPLYSEARCHC );
  ATTATCHSTATUSPTR( status );


  /*
   *
   * check input, output and parameters
   *
   */


  ASSERT( output, status, APPLYSEARCH_ENULL, APPLYSEARCH_MSGENULL );
  ASSERT( !(output->result), status, APPLYSEARCH_ENNUL, APPLYSEARCH_MSGENNUL );

  ASSERT( applyParams, status, APPLYSEARCH_ENULL, APPLYSEARCH_MSGENULL );
  ASSERT( applyParams->mpiParams, status, 
      APPLYSEARCH_ENULL, APPLYSEARCH_MSGENULL );
  ASSERT( applyParams->searchParams, status, 
      APPLYSEARCH_ENULL, APPLYSEARCH_MSGENULL );

  params = (InspiralSearchParams *) applyParams->searchParams;


  /*
   *
   * set up or update the mpi communications within the search comm
   *
   */


  /* if search master, get the initial number of nodes in the serach comm */
  if ( params->searchMaster )
  {
    /* only do this the first time the master is called */
    if ( !params->numSlaves )
    {
      /* allocate memory and check for success */
      params->numSlaves = (UINT4 *) LALMalloc( sizeof(UINT4) );
      if( !params->numSlaves )
      {
        ABORT( status, APPLYSEARCH_EALOC, APPLYSEARCH_MSGEALOC );
      }
      memset( params->numSlaves, 0, sizeof(UINT4) );

      /* get the number of search slaves */
      MPI_Comm_size( *(applyParams->mpiParams->comm), params->numSlaves );
      *(params->numSlaves) -= 1;
    }
  }

  /* get my rank */
  MPI_Comm_rank( *(applyParams->mpiParams->comm), &myRank );

  /* check that the master is running on node rank zero */
  if ( params->searchMaster && myRank != 0 )
  {
    ABORT( status, APPLYSEARCH_ERANK, APPLYSEARCH_MSGERANK );
  }

  /* and that the slave is not running on node rank zero */
  if ( !params->searchMaster && myRank == 0 )
  {
    ABORT( status, APPLYSEARCH_ERANK, APPLYSEARCH_MSGERANK );
  }
    

  /*
   *
   * set up the internal parameters structures
   *
   */


  /* serach master */
  if ( params->searchMaster )
  {
    masterParams = 
      (FindChirpMasterParams *) LALMalloc( sizeof(FindChirpMasterParams) );
    if ( !masterParams )
    {
      ABORT( status, APPLYSEARCH_EALOC, APPLYSEARCH_MSGEALOC );
    }
    memset( masterParams, 0, sizeof(FindChirpMasterParams) );

    masterParams->inspiralDebugFlag = params->inspiralDebugFlag;
    masterParams->numTmpltExch      = params->numTmpltExch;
    masterParams->mpiComm           = applyParams->mpiParams->comm;
    masterParams->numSlaves         = params->numSlaves;
    masterParams->numTmplts         = params->numTmplts;
    masterParams->tmpltCurrent      = params->tmpltCurrent;
    masterParams->tmpltHead         = params->tmpltHead;
    masterParams->notFinished       = &(output->notFinished);
    masterParams->fracRemaining     = &(output->fracRemaining);
    /* these are parameters used for debugging */
    masterParams->loudestEvent      = params->loudestEvent;
  }
  else /* search slave */
  {
    slaveParams = 
      (FindChirpSlaveParams *) LALMalloc (sizeof(FindChirpSlaveParams));
    if ( !slaveParams )
    {
      ABORT( status, APPLYSEARCH_EALOC, APPLYSEARCH_MSGEALOC );
    }
    memset( slaveParams, 0, sizeof(FindChirpSlaveParams) );

    slaveParams->inspiralDebugFlag = params->inspiralDebugFlag;
    slaveParams->mpiComm           = applyParams->mpiParams->comm;
    slaveParams->rhosqThreshVec    = params->rhosqThreshVec;
    slaveParams->chisqThreshVec    = params->chisqThreshVec;
    slaveParams->tmpltParams       = params->tmpltParams;
    slaveParams->filterParams      = params->filterParams;
    slaveParams->filterInput       = params->filterInput;
    /* these are parameters used for debugging */
    slaveParams->loudestEvent      = params->loudestEvent;
  } /* end if master/slave */


  /*
   *
   * execute the appropriate master or slave engine
   *
   */


  if ( params->searchMaster )
  {
    /* set the recieved event list to NULL */
    eventList = NULL;

    /* call the inspiral master engine: this will return with           */
    /* output->notFinished and output->fracRemaining set appropriately  */
    /* and a linked list of InpiralEvents (NULL if no events received   */
    /* it also set masterParams->currentTmplt, but params->currentTmplt */
    /* must be updated manually afterwards                              */
    LALFindChirpMaster (status->statusPtr, &eventList, masterParams);
    CHECKSTATUSPTR (status);

    /* store the current template in the params for the next call */
    params->tmpltCurrent = masterParams->tmpltCurrent;

    /* if events are returned, generate the outPut structure */
    if ( eventList )
    {
      /* create outPut */
      BuildInspiralOutPut (status->statusPtr, output, eventList);
      CHECKSTATUSPTR (status);
    }
    else
    {
      output->numOutput     = 0;
      output->result        = NULL;
    }

    /* if not NULL, free the linked list of InspiralEvents */
    while ( eventList )
    {
      InspiralEvent *thisEvent;
      thisEvent = eventList;
      eventList = eventList->next;
      LALFree (thisEvent);
    }

    /* clear the masterParams */
    LALFree (masterParams);

  }
  else /* search slave */
  {
    /* call the inspiral slave engine: this will return with            */
    /* output->notFinished and output->fracRemaining set appropriately  */
    LALFindChirpSlave (status->statusPtr, &(output->notFinished), 
        params->fcSegVec, slaveParams);
    CHECKSTATUSPTR (status);

    /* if the filter output was requested and this is not the last loop */
    /* through the slave, store the snrsq vector in an output strcuture */
    if ( params->filterParams->rhosqVec && output->notFinished )
    {
      BuildFilterOutPut( status->statusPtr, output, 
          params->filterParams->rhosqVec );
      CHECKSTATUSPTR( status );
    }
    else
    {
    /* search slave does not return results to the wrapper*/
    output->numOutput     = 0;
    output->result        = NULL;
    }
    
    /* set the not finished flag. this should be ignored */
    /* by the wrapper on the serach slaves               */
    if ( output->notFinished )
    {
      output->fracRemaining = 1.0;
    }
    else
    {
      output->fracRemaining = 0;
    }

    /* free slave parameter structure */
    LALFree (slaveParams);

  } /* end if master/slave */


  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}

