#if 0  /* autodoc block */

<lalVerbatim file="SickCV">
$Id: Sick.c,v 1.10 2001/08/29 17:19:30 jolien Exp $
</lalVerbatim>

<lalLaTeX>
\subsection{Module \texttt{Sick.c}}
\label{ss:Sick.c}

The LAL-wrapper interface functions for the \texttt{sick} shared object.

\subsubsection*{Description}

The \verb+filterparams+ input should be a single integer representing the
illness.  The valid illnesses are:
\begin{tabular}{lll}
  0 & ILL\_OK             & OK \\
  1 & ILL\_FAIL           & Failed LAL routine \\
  2 & ILL\_RECURSE\_FAIL  & Recursively failed LAL routine \\
  3 & ILL\_STAT\_DESCRIPT & Status descriptio too long \\
  4 & ILL\_STAT\_LIST     & Too many levels in status linked list \\
  5 & ILL\_LEAK           & Memory leak \\
  6 & ILL\_OUTPUT\_OK     & Fine output to be freed \\
  7 & ILL\_OUTPUT\_PTR    & Output with dodgy pointer \\
  8 & ILL\_OUTPUT\_LIST   & Output with dodgy link to middle of list \\
  9 & ILL\_OUTPUT\_TANGLE & Tangled output \\
 10 & ILL\_SIGSEGV\_INIT  & SIGSEGV in initSearch \\
 11 & ILL\_SIGABRT\_INIT  & SIGABRT in initSearch \\
 12 & ILL\_SIGSEGV\_COND  & SIGSEGV in conditionData \\
 13 & ILL\_SIGABRT\_COND  & SIGABRT in conditionData \\
 14 & ILL\_SIGSEGV\_APPL  & SIGSEGV in applySearch \\
 15 & ILL\_SIGABRT\_APPL  & SIGABRT in applySearch \\
 16 & ILL\_SIGSEGV\_FREE  & SIGSEGV in finalizeSearch \\
 17 & ILL\_SIGABRT\_FREE  & SIGABRT in finalizeSearch \\
 18 & ILL\_INT            & SIGINT somewhere 
\end{tabular}

\subsubsection*{Algorithm}

\subsubsection*{Uses}

\subsubsection*{Notes}

\vfill{\footnotesize\input{SickCV}}

</lalLaTeX>

#endif /* autodoc block */


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <lal/LALStdlib.h>
#include <Sick.h>

#define STR( s ) #s
#define ILLCODE( s ) printf( "Illness code " STR( s ) " = %d\n", s )

NRCSID( SICKC, "$Id: Sick.c,v 1.10 2001/08/29 17:19:30 jolien Exp $" ); 

static void LALRecursiveFailFunction( LALStatus *status, UINT4 level );

void LALInitSearch(
    LALStatus             *status,
    void                 **searchParams,
    LALInitSearchParams   *initSearchParams
    )
{
  int *illness;

  INITSTATUS( status, "LALInitSearch", SICKC );

  ASSERT(    searchParams,  status, SICKH_ENULL, SICKH_MSGENULL );
  ASSERT( ! *searchParams,  status, SICKH_ENNUL, SICKH_MSGENNUL );
  ASSERT( initSearchParams, status, SICKH_ENULL, SICKH_MSGENULL );

  /* use malloc here so that this won't interfere with tests */
  *searchParams = illness = malloc( sizeof( *illness ) );
  if ( ! illness )
  {
    ABORT( status, SICKH_EALOC, SICKH_MSGEALOC );
  }

  switch ( initSearchParams->argc )
  {
    case 1: /* no arguments, no illness */
      *illness = ILL_OK;
      break;
    case 2: /* illness is argument */
      *illness = atoi( initSearchParams->argv[1] );
      break;
    default:
      ABORT( status, SICKH_EARGS, SICKH_MSGEARGS );

  }

  printf( "using illness code: %d\n", *illness );

  if ( *illness == ILL_FAIL ) /* simply fail */
  {
    ILLCODE( ILL_FAIL );
    ABORT( status, SICKH_EFAIL, SICKH_MSGEFAIL );
  }

  if ( *illness == ILL_RECURSE_FAIL ) /* fail from a recursive function */
  {
    ILLCODE( ILL_RECURSE_FAIL );
    ATTATCHSTATUSPTR( status );
    LALRecursiveFailFunction( status->statusPtr, 1 );
    CHECKSTATUSPTR( status );
    DETATCHSTATUSPTR( status );
  }

  if ( *illness == ILL_STAT_DESCRIPT ) /* bad status description */
  {
    char statusDescription[32767];
    ILLCODE( ILL_STAT_DESCRIPT );
    memset( statusDescription, '?', 32765 );
    statusDescription[16386] = '\0';
    SETSTATUSFILELINE( status );
    status->statusCode = SICKH_EFAIL;
    status->statusDescription = (const char *)statusDescription;
    LALError( status, "ABORT:" );
    return;
  }

  if ( *illness == ILL_STAT_LIST ) /* bad status linked list */
  {
    ILLCODE( ILL_STAT_LIST );
    ATTATCHSTATUSPTR( status );
    LALRecursiveFailFunction( status->statusPtr, 65 );
    CHECKSTATUSPTR( status );
    DETATCHSTATUSPTR( status );
  }

  if ( *illness == ILL_SIGSEGV_INIT )
  {
    ILLCODE( ILL_SIGSEGV_INIT );
    raise( SIGSEGV );
  }

  if ( *illness == ILL_SIGABRT_INIT )
  {
    ILLCODE( ILL_SIGABRT_INIT );
    raise( SIGABRT );
  }

  if ( *illness == ILL_LEAK )
  {
    char *lost = LALMalloc( 1 );
    lost = "all gone, goodbye!";
    ILLCODE( ILL_LEAK );
  }

  if ( *illness == ILL_INT )
  {
    ILLCODE( ILL_INT );
    if ( initSearchParams->nodeClass != 1 )
      raise( SIGINT );
  }

  RETURN( status );
}


void
LALConditionData(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams          *mpiParams
    )
{
  int *illness = searchParams;

  INITSTATUS( status, "LALConditionData", SICKC );

  ASSERT( inout,   status, SICKH_ENULL, SICKH_MSGENULL );
  ASSERT( illness, status, SICKH_ENULL, SICKH_MSGENULL );
  mpiParams = NULL;

  if ( *illness == ILL_SIGSEGV_COND )
  {
    ILLCODE( ILL_SIGSEGV_COND );
    raise( SIGSEGV );
  }

  if ( *illness == ILL_SIGABRT_COND )
  {
    ILLCODE( ILL_SIGABRT_COND );
    raise( SIGABRT );
  }

  RETURN( status );
}

void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *params
    )
{
  int *illness;

  INITSTATUS( status, "LALApplySearch", SICKC );
  
  ASSERT( output, status, SICKH_ENULL, SICKH_MSGENULL );
  ASSERT( input,  status, SICKH_ENULL, SICKH_MSGENULL );
  ASSERT( params, status, SICKH_ENULL, SICKH_MSGENULL );

  illness = params->searchParams;
  ASSERT( illness, status, SICKH_ENULL, SICKH_MSGENULL );

  if ( *illness == ILL_SIGSEGV_APPL )
  {
    ILLCODE( ILL_SIGSEGV_APPL );
    raise( SIGSEGV );
  }

  if ( *illness == ILL_SIGABRT_APPL )
  {
    ILLCODE( ILL_SIGABRT_APPL );
    raise( SIGABRT );
  }

  switch ( *illness )
  {
    case ILL_OK: /* no illness, no output */
        output->numOutput = 0;
        output->result    = NULL;
        break;
    case ILL_OUTPUT_OK:
        ILLCODE( ILL_OUTPUT_OK );
        output->numOutput = 2;
        output->result    = LALCalloc( 2, sizeof( *output->result ) );
        output->result->optional =
          LALCalloc( 1, sizeof( *output->result->optional ) );
        output->result->optional->history =
          LALCalloc( 1, sizeof( *output->result->optional->history ) );
        output->result->optional->history->next =
          LALCalloc( 1, sizeof( *output->result->optional->history->next ) );
        break;
    case ILL_OUTPUT_PTR:
        ILLCODE( ILL_OUTPUT_PTR );
        output->numOutput = 1;
        output->result    = (void *)64;
        break;
    case ILL_OUTPUT_LIST:
        ILLCODE( ILL_OUTPUT_LIST );
        output->numOutput = 1;
        output->result    = LALCalloc( 1, sizeof( *output->result ) );
        output->result->optional =
          LALCalloc( 1, sizeof( *output->result->optional ) );
        output->result->optional->history =
          LALCalloc( 1, sizeof( *output->result->optional->history ) );
        output->result->optional->history->previous =
          LALCalloc( 1, sizeof( *output->result->optional->history->previous ) );
        output->result->optional->history->next =
          LALCalloc( 1, sizeof( *output->result->optional->history->next ) );
        break;
    case ILL_OUTPUT_TANGLE:
        ILLCODE( ILL_OUTPUT_TANGLE );
        output->numOutput = 1;
        output->result    = LALCalloc( 1, sizeof( *output->result ) );
        output->result->states =
          LALCalloc( 1, sizeof( *output->result->states ) );
        output->result->states->store =
          LALCalloc( 1, sizeof( *output->result->states->store ) );
        output->result->optional = output->result->states->store;
        break;
  }
  output->fracRemaining = 0;
  output->notFinished   = 0;

  RETURN( status );
}

void
LALFinalizeSearch(
    LALStatus             *status,
    void                 **searchParams
    )
{
  int *illness;

  INITSTATUS( status, "LALFinalizeSearch", SICKC );

  ASSERT(  searchParams, status, SICKH_ENULL, SICKH_MSGENULL );
  ASSERT( *searchParams, status, SICKH_ENULL, SICKH_MSGENULL );

  illness = *searchParams;

  if ( *illness == ILL_SIGSEGV_FREE )
  {
    ILLCODE( ILL_SIGSEGV_FREE );
    raise( SIGSEGV );
  }

  if ( *illness == ILL_SIGABRT_FREE )
  {
    ILLCODE( ILL_SIGABRT_FREE );
    raise( SIGABRT );
  }

  /* this was allocated with malloc, so just free it */
  free( *searchParams );
  *searchParams = NULL;

  RETURN( status );
}

static void LALRecursiveFailFunction( LALStatus *status, UINT4 level )
{
  INITSTATUS( status, "LALRecursiveFailFunction", SICKC );
  if ( level-- )
  {
    ATTATCHSTATUSPTR( status );
    LALRecursiveFailFunction( status->statusPtr, level );
    CHECKSTATUSPTR( status );
    DETATCHSTATUSPTR( status );
  }
  else
  {
    ABORT( status, SICKH_EFAIL, SICKH_MSGEFAIL );
  }
  RETURN( status );
}
