#if 0  /* autodoc block */

<lalVerbatim file="LALWrapperInterfaceCV">
$Id: LALWrapperInterface.c,v 1.9 2001/08/27 16:40:45 jolien Exp $
</lalVerbatim>

<lalLaTeX>

\subsection{Module \texttt{LALWrapperInterface.c}}
\label{ss:LALWrapperInterface.c}

This module provides the routines called by \texttt{wrapperAPI} which ``wrap''
the generic LAL routines.  This module, therefore, is not search specific.

The general tasks that this module perfoms are:
\begin{itemize}
\item Defines \verb+lalDebugLevel+.
\item Defines a \verb+LALStatus+ variable for each of the LAL-Wrapper interface
      generic functions.
\item Defines a global variable that contains specific search parameters,
      which is created by \verb+LALInitSearch()+ and destroyed by
      \verb+LALFinalizeSearch()+, and is passed to \verb+LALApplySearch()+
      and \verb+LALConditionData()+.
\item Re-packages variables given by \verb+wrapperAPI+ into the structures
      used by the LAL-Wrapper interface generic functions.
\item Parses the exit status from the LAL-Wrapper interface generic functions
      into a character string used by \verb+wrapperAPI+.
\item Provides a routine for freeing output memory declared within the search
      specific LAL routines.
\item Handles various signals that may be raised by LAL.
\end{itemize}

\vfill{\footnotesize\input{LALWrapperInterfaceCV}}

</lalLaTeX>

#endif /* autodoc block */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <wrapperInterface.h>
#include <LALWrapperInterface.h>
#include <lal/LALStdlib.h>


NRCSID( LALWRAPPERINTERFACEC, "$Id: LALWrapperInterface.c,v 1.9 2001/08/27 16:40:45 jolien Exp $" );
#define STRINGIFY( a ) #a
#define XSTRINGIFY( a ) STRINGIFY( a )
#define FILELINEID ", file: " __FILE__                   \
                   ", line: " XSTRINGIFY( __LINE__ )     \
                   ", $Id: LALWrapperInterface.c,v 1.9 2001/08/27 16:40:45 jolien Exp $"


/*
 *
 * wrapperAPI will call functions at debuglevel = 0
 *
 */

#ifdef NDEBUG
int lalDebugLevel = LALNDEBUG;
#else
int lalDebugLevel = LALMSGLVL3;
/* int debuglevel = LALALLDBG; */
#endif

enum { Warning = -1, Nominal = 0, Error = 1 }; /* possible return codes */

static void clearStateVector( stateVector *state );
static void clearDataBase( dataBase *dbase );
static void clearMultiDimData( multiDimData *data );
static void clearDCHistory( dcHistory *history );
static void signalHandler( int sig );
static INT4 stringifyLALStatus( CHAR **msgString, LALStatus *status );


/*
 *
 * Global variables.
 *
 */


static BOOLEAN  wrapperMaster    = 0;
static void    *searchParams     = NULL;
static CHAR    *exceptionMessage = NULL;
static jmp_buf  jump;


/*
 *
 * Function initSearch()
 * Function conditionData()
 * Function applySearch()
 * Function finalizeSearch()
 * Function freeOutput()
 *
 * These are the functions defined in the Wrapper API's baseline requirements.
 *
 */


INT4 initSearch( CHAR **initStatus, InitParams *initParams )
{
  struct sigaction action        = {};
  struct sigaction oldABRTAction = {};
  struct sigaction oldSEGVAction = {};
  INT4 code;

  /* initStatus must be a pointer to NULL */
  if ( !initStatus || *initStatus )
  {
    return Error;
  }

  /* initParams must be a non-NULL pointer */
  if ( !initParams )
  {
    char *msg = "initSearch: NULL parameters" FILELINEID;
    *initStatus = realloc( *initStatus, strlen( msg ) + 1 );
    if ( *initStatus )
    {
      strcpy( *initStatus, msg );
    }
    return Error;
  }

  /* initialize custom handler */
  action.sa_handler = signalHandler;
  sigaction( SIGABRT, &action, &oldABRTAction );
  sigaction( SIGSEGV, &action, &oldSEGVAction );

  /* try InitFilters; catch SIGABRT and SIGSEGV with custom handler */
  if ( setjmp( jump ) == 0 )
  {
    LALStatus status = { 0 };
    LALInitSearch( &status, &searchParams, initParams );
    code = stringifyLALStatus( initStatus, &status );
  }
  else
  {
    *initStatus = realloc( *initStatus, strlen( exceptionMessage ) + 1 );
    if ( *initStatus )
    {
      strcpy( *initStatus, exceptionMessage );
    }
    code = Error;
  }

  /* restore previous handlers */
  sigaction( SIGABRT, &oldABRTAction, NULL );
  sigaction( SIGSEGV, &oldSEGVAction, NULL );

  return code;
}


INT4 conditionData( CHAR **conditionStatus, inPut *data,
    SearchParams* mpiCommParams )
{
  struct sigaction action        = {};
  struct sigaction oldABRTAction = {};
  struct sigaction oldSEGVAction = {};
  INT4 code;

  /* conditionMessage must be a pointer to NULL */
  if ( !conditionStatus || *conditionStatus )
  {
    return Error;
  }

  /* just return if I am the wrapper master */
  if ( wrapperMaster )
  {
    return Nominal;
  }

  /* initialize custom handler */
  action.sa_handler = signalHandler;
  sigaction( SIGABRT, &action, &oldABRTAction );
  sigaction( SIGSEGV, &action, &oldSEGVAction );

  /* try ConditionData; catch SIGABRT and SIGSEGV with custom handler */
  if ( setjmp( jump ) == 0 )
  {
    LALStatus status = { 0 };
    LALConditionData( &status, data, searchParams, mpiCommParams );
    code = stringifyLALStatus( conditionStatus, &status );
  }
  else
  {
    *conditionStatus = realloc( *conditionStatus,
                                 strlen( exceptionMessage ) + 1 );
    if ( *conditionStatus )
    {
      strcpy( *conditionStatus, exceptionMessage );
    }
    code = Error;
  }

  /* restore previous handlers */
  sigaction( SIGABRT, &oldABRTAction, NULL );
  sigaction( SIGSEGV, &oldSEGVAction, NULL );

  return code;
}


INT4 applySearch( CHAR **searchStatus, inPut *input, SearchOutput *output,
                  SearchParams *mpiParams )
{
  struct sigaction action        = {};
  struct sigaction oldABRTAction = {};
  struct sigaction oldSEGVAction = {};
  INT4 code;

  /* filterMessage must be a pointer to NULL */
  if ( !searchStatus || *searchStatus )
  {
    return Error;
  }

  /* just return if I am the wrapper master */
  if ( wrapperMaster )
  {
    return Nominal;
  }

  /* initialize custom handler */
  action.sa_handler = signalHandler;
  sigaction( SIGABRT, &action, &oldABRTAction );
  sigaction( SIGSEGV, &action, &oldSEGVAction );

  /* try TemplateFilters; catch SIGABRT and SIGSEGV with custom handler */
  if ( setjmp( jump ) == 0 )
  {
    LALStatus            status = { 0 };
    LALApplySearchParams params = { 0 };
    params.mpiParams    = mpiParams;
    params.searchParams = searchParams;
    LALApplySearch( &status, output, input, &params );
    code = stringifyLALStatus( searchStatus, &status );
  }
  else
  {
    *searchStatus = realloc( *searchStatus, strlen( exceptionMessage ) + 1 );
    if ( *searchStatus )
    {
      strcpy( *searchStatus, exceptionMessage );
    }
    code = Error;
  }

  /* restore previous handlers */
  sigaction( SIGABRT, &oldABRTAction, NULL );
  sigaction( SIGSEGV, &oldSEGVAction, NULL );

  return code;
}


INT4 finalizeSearch( CHAR **finalizeStatus )
{
  struct sigaction action        = {};
  struct sigaction oldABRTAction = {};
  struct sigaction oldSEGVAction = {};
  INT4 code;

  /* freeMessage must be a pointer to NULL */
  if ( !finalizeStatus || *finalizeStatus )
  {
    return Error;
  }

  /* just return if I am the wrapper master */
  if ( wrapperMaster )
  {
    return Nominal;
  }

  /* initialize custom handler */
  action.sa_handler = signalHandler;
  sigaction( SIGABRT, &action, &oldABRTAction );
  sigaction( SIGSEGV, &action, &oldSEGVAction );

  /* try FreeFilters; catch SIGABRT and SIGSEGV with custom handler */
  if ( setjmp( jump ) == 0 )
  {
    LALStatus status = { 0 };
    LALFinalizeSearch( &status, &searchParams );
    code = stringifyLALStatus( finalizeStatus, &status );
    if ( code != Error )
    {
      LALCheckMemoryLeaks();
    }
  }
  else
  {
    *finalizeStatus = realloc( *finalizeStatus,
                               strlen( exceptionMessage ) + 1 );
    if ( *finalizeStatus )
    {
      strcpy( *finalizeStatus, exceptionMessage );
    }
    code = Error;
    goto exit;
  }

exit:

  /* restore previous handlers */
  sigaction( SIGABRT, &oldABRTAction, NULL );
  sigaction( SIGSEGV, &oldSEGVAction, NULL );

  return code;
}


INT4 freeOutput( CHAR **freeStatus, SearchOutput *output )
{
  struct sigaction action        = {};
  struct sigaction oldABRTAction = {};
  struct sigaction oldSEGVAction = {};
  INT4 code = Nominal;
  INT4 i;

  /* freeMessage must be a pointer to NULL */
  if ( !freeStatus || *freeStatus )
  {
    return Error;
  }

  /* just return if I am the wrapper master */
  if ( wrapperMaster )
  {
    return Nominal;
  }

  /* initialize custom handler */
  action.sa_handler = signalHandler;
  sigaction( SIGABRT, &action, &oldABRTAction );
  sigaction( SIGSEGV, &action, &oldSEGVAction );

  /* try to free memory; catch SIGABRT and SIGSEGV with custom handler */
  if ( setjmp( jump ) == 0 )
  {
    code = Nominal;
    if ( output )
    {
      if ( output->result )
      {
        for ( i = 0; i < output->numOutput; ++i )
        {
          if ( output->result[i].states )
          {
            clearStateVector( output->result[i].states );
            output->result[i].states = NULL;
          }
          if ( output->result[i].results )
          {
            clearDataBase( output->result[i].results );
            output->result[i].results = NULL;
          }
          if ( output->result[i].optional )
          {
            clearMultiDimData( output->result[i].optional );
            output->result[i].optional = NULL;
          }
        }
        LALFree( output->result );
        output->result = NULL;
      }
      memset( output, 0, sizeof( *output ) );
    }
    else
    {
      CHAR *message = "Null Pointer to Search Output.";
      *freeStatus = realloc( *freeStatus, strlen( message ) + 1 );
      if ( *freeStatus )
      {
        strcpy( *freeStatus, message );
      }
      code = Error;
    }
  }
  else
  {
    *freeStatus = realloc( *freeStatus, strlen( exceptionMessage ) + 1 );
    if ( *freeStatus )
    {
      strcpy( *freeStatus, exceptionMessage );
    }
    code = Error;
  }

  /* restore previous handlers */
  sigaction( SIGABRT, &oldABRTAction, NULL );
  sigaction( SIGSEGV, &oldSEGVAction, NULL );

  return code;
}

/*
 *
 * Function clearStateVector()
 * Function clearDataBase()
 * Function clearMultiDimData()
 * Function clearDCHistory()
 *
 * Routines to clear various data types, recursively freeing allocated memory.
 *
 */

static void clearStateVector( stateVector *state )
{
  if ( state )
  {
    if ( state->previous )
    {
      state->previous->next = NULL; /* limit recursion */
      clearStateVector( state->previous );
      state->previous = NULL;
    }
    if ( state->next )
    {
      state->next->previous = NULL; /* limit recursion */
      clearStateVector( state->next );
      state->next = NULL;
    }
    if ( state->store )
    {
      clearMultiDimData( state->store );
      state->store = NULL;
    }
    memset( state, 0, sizeof( *state ) );
    LALFree( state );
  }
  return;
}

static void clearDataBase( dataBase *dbase )
{
  if ( dbase )
  {
    if ( dbase->previous )
    {
      dbase->previous->next = NULL; /* limit recursion */
      clearDataBase( dbase->previous );
      dbase->previous = NULL;
    }
    if ( dbase->next )
    {
      dbase->next->previous = NULL; /* limit recursion */
      clearDataBase( dbase->next );
      dbase->next = NULL;
    }
    if ( dbase->rows.chars )
    {
      LALFree( dbase->rows.chars ); /* frees any kind of data pointer */
      dbase->rows.chars = NULL;
      dbase->numberRows = 0;
    }
    if ( dbase->rowDimensions )
    {
      LALFree( dbase->rowDimensions ); /* frees BLOB row dimensions */
      dbase->rowDimensions = NULL;
    }
    memset( dbase, 0, sizeof( *dbase ) );
    LALFree( dbase );
  }
  return;
}

static void clearMultiDimData( multiDimData *data )
{
  if ( data )
  {
    if ( data->dimensions )
    {
      LALFree( data->dimensions );
      data->dimensions = NULL;
      data->numberDimensions = 0;
    }
    if ( data->history )
    {
      clearDCHistory( data->history );
      data->history = NULL;
    }
    if ( data->data.chars )
    {
      LALFree( data->data.chars );
      data->data.chars = NULL;
    }
    memset( data, 0, sizeof( *data ) );
    LALFree( data );
  }
  return;
}

static void clearDCHistory( dcHistory *history )
{
  if ( history )
  {
    if ( history->previous )
    {
      history->previous->next = NULL; /* limit recursion */
      clearDCHistory( history->previous );
      history->previous = NULL;
    }
    if ( history->next )
    {
      history->next->previous = NULL; /* limit recursion */
      clearDCHistory( history->next );
      history->next = NULL;
    }
    if ( history->value.chars )
    {
      LALFree( history->value.chars );
      history->value.chars  = NULL;
      history->numberValues = 0;
    }
    memset( history, 0, sizeof( *history ) );
    LALFree( history );
  }
  return;
}


/*
 *
 * Function signalHandler()
 *
 * Handles exceptions SIGABRT and SIGSEGV.  Error message is written to
 * exceptionMessage.
 *
 */
static void signalHandler( int sig )
{
  switch ( sig )
  {
    case SIGABRT:
      exceptionMessage = "signalHandler: Caught SIGABRT" FILELINEID;
      break;
    case SIGSEGV:
      exceptionMessage = "signalHandler: Caught SIGSEGV" FILELINEID;
      break;
    default:
      exceptionMessage = "signalHandler: Caught unknown signal" FILELINEID;
      break;
  }
  longjmp( jump, sig );
}


/*
 *
 * Function stringifyLALStatus()
 *
 * Converts the status structure linked list into a string *msgString,
 * for which memory is allocated.  Meanwhile, frees up memory of the linked
 * list of status structures.  Return values are:
 *
 * -1 (Warning) if the top-level statusCode was non-zero (abnormal termination)
 *  0 (Nominal) if the top-level statusCode was zero (normal termination)
 * +1 (Error)   if there is an internal error in this routine
 *
 */
static INT4 stringifyLALStatus( CHAR **msgString, LALStatus *status )
{
  enum { MaxNumLevels  = 64    }; /* maximum levels in status list           */
  enum { MsgStringSize = 16384 }; /* size of the message string              */

  LALStatus *ptr           = status;
  INT4       code          = Nominal;
  unsigned   level         = 0;
  int        numCharRemain = MsgStringSize;
  CHAR      *msgStringEnd;  


  /*
   *
   * If msgString is NULL, or we can't allocate memory, we're hozed!
   * Also check to see if status is null.
   *
   */


  if ( ! msgString || *msgString )
  {
    return Error;
  }

  msgStringEnd  = *msgString = calloc( MsgStringSize, sizeof( CHAR ) );
  if ( ! *msgString )
  {
    return Error;
  }

  if ( !status )
  {
    memcpy( *msgString, "stringifyLALStatus: null status structure" FILELINEID,
        MsgStringSize );
    return Error;
  }


  /*
   *
   * Determine exit code.
   *
   */


  code = status->statusCode ? Error : Nominal ;


  /*
   *
   * Loop over levels of the status structure linked list until end of list
   * is found.  At each level, write information onto the message string
   * msgString.
   *
   */


  while ( ptr )
  {
    LALStatus *next = ptr->statusPtr;
    int        numChar;


    /*
     *
     * If level is greater than MaxNumLevels there is probably a serious
     * error somewhere....
     *
     */


    if ( ++level > MaxNumLevels )
    {
      memcpy( *msgString, "stringifyLALStatus: too many levels in status "
          "structure" FILELINEID, MsgStringSize );
      return Error;
    }


    /*
     *
     * Write this level of status information to tmpString.
     * Keep track of total number of characters written (numChar),
     * and abort if there is an error in snprintf() or if numChar
     * exceeds storage size of tmpString.
     *
     */


    if ( ptr->statusCode ) /* non-zero status code: an error */
    {
      numChar = snprintf( msgStringEnd, numCharRemain, "\nLevel %i: %s\n"
          "\tStatus code %i: %s\n\tfunction %s, file %s, line %i\n",
          ptr->level, ptr->Id, ptr->statusCode, ptr->statusDescription,
          ptr->function, ptr->file, ptr->line );
    }
    else /* zero status code: nominal */
    {
      numChar = snprintf( msgStringEnd, numCharRemain, "\nLevel %i: %s\n"
          "\tStatus code 0: Nominal\n\tfunction %s, file %s, line %i\n",
          ptr->level, ptr->Id, ptr->function, ptr->file, ptr->line );
    }

    if ( numChar < 0 )
    {
      memcpy( *msgString, "stringifyLALStatus: an error in snprintf()"
          FILELINEID, MsgStringSize );
      return Error;  /* an error in snprintf() */
    }

    msgStringEnd  += numChar;
    numCharRemain -= numChar;

    if ( numCharRemain <= 0 )
    {
      memcpy( *msgString, "stringifyLALStatus: error message too long"
          FILELINEID, MsgStringSize );
      return Error;  /* error message is too long */
    }


    /*
     *
     * Deallocate the memory of this level of the status linked list unless
     * this is the top level; connect ptr to the next level of the list.
     *
     */


    if ( ptr != status )
    {
      LALFree( ptr );
    }
    ptr = next;

  } /* while ( ptr ) */

  return code;
}


#undef STRINGIFY_HELPER
#undef STRINGIFY
#undef FILELINEID
